/*
 * Test / Demo for the non-Thing system use of things
 *
 * The first time the program is run, it create a shared data thing
 * called "Test Stuff".
 *
 * The second time it's run, it USEs "Test Stuff", then creates a child
 * job that modifies the shared data,
 * and then tidies up, removing the shared thing.
 *
 * DO NOT RUN THIS with THING System installed, it will screw it up, it
 * just illustrates a way of adding THING type shared data even if you don't
 * have the THING system.
 */

#include <stdio.h>
#include <stdlib.h>
#include <qdos.h>
#include <qlib.h>
#include <pd/thing.h>

extern char *_sys_var;

/*
 * We use this structure to describe the our shared data thing
 * the flag and type fields are mandatory QJump fields. We have
 * 32 bytes for our data (this is a trivial example :-)
 */

typedef struct
{
    int  flag;
    int  type;
    char text[32];
} TEST_DATA, *PTEST_DATA;


#define TEST_ID  0x302e3031         /* 0.01 */

/*
 * Let's just have one of these
 */
 
static char *Stuff = "Test Stuff";

/*
 * Here we define our simple thing
 */
 
PTHG_LINK AddThing(char *name, char *txt, PTHG_LINK *p)
{
    int m,n;
    int size;
    PTHG_LINK q;
    
    n = strlen(name);

    /*
     * size = size of a linkage block plus the length of the name, and is
     * rounded to the next even number if necessary
     */
     
    size = sizeof(THG_LINK) + ((n + 1) & ~1);

    /*
     * Make one allocation for both linkage block and data
     * Owned by job zero
     */
     
    *p = (PTHG_LINK) thg_malloc ((m = (size + sizeof(TEST_DATA))), 0);

    printf("Length is %d, size %d request %d\n", n ,size, m);

    if(*p)
    {
        /*
         * Got some memory, so define our thing. We have to set the id
         * the name (and length) and thing address in the linkage block
         * and the header info in the thing itself. We also preset the
         * data in the thing.
         */
         
        (*p)->namelen = n;
        memcpy(((*p)+1), name, n);
        (*p)->verid = TEST_ID;

        /*
         * This is where the thing starts.
         */
         
        (*p)->thing = ((char *)(*p) + size);

        ((PTEST_DATA) (*p)->thing)->flag = THH_FLAG;
        ((PTEST_DATA) (*p)->thing)->type = THT_DATA;
        strcpy( ((PTEST_DATA) (*p)->thing)->text, txt);

        /*
         * Here we link the thing into the system list
         * Don't do this if you have the THING system, or
         * you confuse it !!!!
         */

        q = (PTHG_LINK)(_sys_var + SYS_THGL);
        (*p)->nxtth = q->nxtth;
        q->nxtth = *p;
    }
}

/*
 * A quick and dirty way to list things without using the thing system
 */
 
list_things()
{
    PTHG_LINK p;
    

    for (p = ((PTHG_LINK)(_sys_var + SYS_THGL))->nxtth; p; p = p->nxtth) 
    {
        printf("%x %x %.*s\n", p, p->nxtth, p->namelen, (char *)(p + 1));
    }    
}

/*
 * A quick and dirty way to find a thing without using the thing system
 */
 
PTHG_LINK find_thing(char *name)
{
    PTHG_LINK p;
    int n;

    n = strlen(name);

    for (p = ((PTHG_LINK)(_sys_var + SYS_THGL))->nxtth; p; p = p->nxtth) 
    {
        if (n == p->namelen && strnicmp(name, (char *)(p + 1), n) == 0)
        {
            break;
        }
    }
    return p;
}
PTHG_LINK remv_thing(char *name)
{
    PTHG_LINK p,q;
    int n;

    n = strlen(name);
    q =  (PTHG_LINK)(_sys_var + SYS_THGL);

    for (p = q->nxtth; p; p = p->nxtth) 
    {
        if (n == p->namelen && strnicmp(name, (char *)(p + 1), n) == 0)
        {
            q->nxtth = p->nxtth;
            thg_free(p);
            break;
        }
        q = p;
    }
    return p;
}

/*
 * Sharer is a little concurrent job that shares our thing. We synchronise
 * usage by using address [31] in the shared data as a flag.
 */ 

int Sharer()
{
    int vers;
    PTHG_LINK p = NULL;    
    void *q;
    int sts;
    
    /*
     * Try to use the thing
     */
     
    p = find_thing(Stuff);
    
    if(p)
    {
        q = p->thing;
        
        /*
         * OK, we can use it, so change the data
         */
        strcpy(((PTEST_DATA)q)->text, "It's a wonderful thing");

        /*
         * Set the flag to show the data has changed
         */

        ((PTEST_DATA)q)->text[31] = 1;

        /*
         * Wait until the parent has done his (her) bit
         */

        while(((PTEST_DATA)q)->text[31])
        {
            mt_susjb(-1, 1, NULL);
        }

        /*
         * Tell the parent job we're done
         */

        ((PTEST_DATA)q)->text[31] = 1;
    }

    return 0;
}

/*
 * The real work starts here
 */

main(int ac, char **av)
{
    PTHG_LINK p = NULL;    
    long vers = 0;
    void *q;
    
    /*
     * See if we can use the thing
     */

    p = find_thing(Stuff);
    
    if(p)
    {
        /* We already have a thing from the last invocation of the
         * program, so let's see if we can (ab)use it.
         */
         
        vers = p->verid;
        q = p->thing;
        
        printf("use: vers is %4.4s addr %x %x\n", &vers, q, p);
        getchar();

        if(q)
        {
            /*
             * That was OK, so show header (THG%) and the type
             * (should be 2) and the data
             */
             
            printf("The text is (%4.4s %d) : %s\n",
                        q, *((int *)q + 1), ((PTEST_DATA)q)->text);

            /*
             * Set the flag so the (to be conceived) child can change
             * the data.
             */
             
            ((PTEST_DATA)q)->text[31] = 0;
        }

        /*
         * Now start the child job
         */
         
	jcall(Sharer, "Share a thing", 16, 1024, 0);

	/*
	 * Idle time while the child changes the thing
	 */
	 
	while(((PTEST_DATA)q)->text[31] == 0)
	{
	    mt_susjb(-1, 1, NULL);
	}

        /*
         * How see what's there
         */
         
	printf("The shared thing is now : %s\n", ((PTEST_DATA)q)->text);

        /*
         * List all things
         */
         
        list_things();        

        /*
         * Show the child it can terminate
         */
         
        ((PTEST_DATA)q)->text[31] = 0;

        /*
         * And wait until its done (ensures we can remove the thing)
         */
         
	while(((PTEST_DATA)q)->text[31] == 0)
	{
	    mt_susjb(-1, 1, NULL);
	}
        fputs("OK ? ", stdout);

        getchar();

        /*
         * Now remove from list
         */

        remv_thing (Stuff);      
    }
    else 
    {
        AddThing (Stuff, "What thing is this ?", &p);
        printf("Add returns %x\n", p);
    }

    list_things();
    return 0;
}
