*
* Hotkey.s 
* 
* A public domain implementation of the HOTKEY II system for c68
* Part of the HOTStuff c68 library
*
* (c) Jonathan Hudson 1993
*

.globl  hk_set
.globl  hk_reset
.globl  hk_on
.globl  hk_off
.globl  hk_fitem
.globl	hk_remv
.globl	hk_sstbf
.globl	hk_getbf
.globl	hk_cjob
.globl	hk_kjob
.globl  hk_do

hk.fitem        equ     $14
hk.crjob	equ	$18
hk.kjob		equ	$1c
hk.set         	equ     $20
hk.remov	equ	$24
hk.do		equ	$28
hk.stbuf	equ	$2c
hk.gtbuf	equ	$30
hk.guard        equ     $34

hks.off 	equ	-1
hks.on		equ	 0
hks.rset       	equ      1
hks.set        	equ      2

sms.uthg 	equ    	$28
sms.fthg 	equ    	$29

err.nimp 	equ    	-19

.text

*
*   Use Hotkey system, but be quick .....
*

hk_use:
        jsr     getthg
	bne	fail
	move.l	a0,a5
	moveq	#sms.uthg,d0
	moveq	#-1,d1
	moveq	#127,d3
	lea.l	hk_thing,a0
	jsr     (a5)
	move.l	a1,a3
fail:
	tst.l	d0
	rts
*
* Free the Hotkey system
*

hk_free:
	moveq	#sms.fthg,d0
	moveq	#-1,d1
	lea.l	hk_thing,a0
	jsr	(a5)
	rts	
*
* hk_set(key, hk_item)
* hk_reset(newkey, kname)
* hk_on (kname)
* hk_off (kname)
*

hk_set:
	moveq	#hks.set,d0         	; action needed
	bra	hkset
hk_reset:
	moveq	#hks.rset,d0         	; action needed
	bra	hkset
hk_on:
	moveq	#hks.on,d0         	; action needed
	bra	hkset
hk_off:
	moveq	#hks.off,d0         	; action needed

hkset:	
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	move.b	d0,d4			; save action
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.b	d4,d0			; what to do
	move.l	hk.set(a3),a4       	; our vector
	cmpi.b	#hks.set,d0		; setting ?
	bne	h1			; no ...
	move.l	8(a6),d1           	; the key
	move.l	12(a6),a1            	; the item address
        move.l  16(a6),a2               ; guard ?
        cmp.l   #0,a2                   ; set
        beq     h4                      ; no
        move.l  hk.guard(a3),2(a2)      ; set vector
	bra	h4			; OK
h1:
	cmpi.b	#hks.rset,d0		; resetting ?
	bne	h2			; no ...
	move.l	12(a6),a1            	; the item address
	jsr	__StackA1		; make descriptor
	move.l	8(a6),d1           	; the key
	bra	h4			; OK
h2:	move.l	8(a6),a1		; must be on/off
	jsr	__StackA1		; stack em up
h4:
	jsr	(a4)                	; do it
        bra     hks_finish

hk_fitem:	
	link	a6,#0               ; use a6 as frame pointer
	movem.l	d2-d4/a2-a5,-(sp)   
	bsr	hk_use              ; get hot_key
	bne	hks_done            ; quit if we cant

	move.l	hk.fitem(a3),a4     ; this is what we want

        move.l	8(a6),a1            ; string
	jsr	__StackA1			
	jsr	(a4)                ; do hk.fitem
        move.l  12(a6),a2           ; item address
        move.l  a1,(a2)             ; return it
        move.l  16(a6),a2           ; hotkey address
        move.w  d1,(a2)             ; return it
        move.l  20(a6),a2           ; hot number address
        move.w  d2,(a2)             ; return it

hks_finish:
	move.l	d0,-(sp)            	; save status
	bsr	hk_free             	; free HK system
	move.l	(sp)+,d0            	; status

hks_done:	
	movem.l	(sp)+,d2-d4/a2-a5
	unlk	a6
	rts

hk_remv:
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.l	hk.remov(a3),a4       	; our vector
        move.l  8(a6),a1                ; item
	jsr	__StackA1
	jsr	(a4)                	; do it
        bra     hks_finish

*
*   hk_sstbf. Stuff into stuffer buffer
*

hk_sstbf:
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.l	hk.stbuf(a3),a4       	; our vector
        move.l  8(a6),-(sp)             ; item
	jsr	__A1len
	jsr	(a4)                	; do it
        bra     hks_finish

*
*	char * hk_gtbuf(key, &ptr)
*

hk_getbf:
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.l	hk.gtbuf(a3),a4       	; our vector
        move.l  8(a6),d0                ; item
	jsr	(a4)                	; do it
        move.l  12(a6),a3               ; return address
        tst.l   d0                      ; OK
        bne     h8                      ; no
	tst.w	d2			; got any
	beq	h8			; no
	addq.w	#1,d2			; plus one for nul
	ext.l	d2			; make long
        move.l  a1,-(sp)                ; save buffer
	move.l	d2,-(sp)		; size
	jsr	malloc			; 'C' malloc
	lea.l	4(sp),sp		; reset stack
        move.l  (sp)+,a1                ; restore buffer
	move.l	d0,(a3) 		; return it
	beq	h9      		; nothing
	subq.w	#2,d2			; less nul, less 1
        move.l  d0,a0                   ; 'C' buffer
        moveq   #0,d0                   ; preset return
h6:
	move.b	(a1)+,(a0)+		; copy data
	dbra	d2,h6			; loop
h7:	move.b	#0,(a0)			; terminate for 'C'			
        bra     hks_finish
h8:
	move.l	#0,(a3)                 ; show NULL
	bra	hks_finish
h9:
        moveq   #-3,d0                  ; no memory
	bra	hks_finish

*
*   hk_do.  Do a hot item
*

hk_do:
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.l	hk.do(a3),a4       	; our vector
	move.l	8(a6),a1
	move.l	a6,-(sp)
	move.l	_SPBASE,a6
	jsr	(a4)
	move.l	(sp)+,a6
	bra	hks_finish

*
*   Create/Kill hot key job
*

hk_cjob:
	move.w	#hk.crjob,d0
	bra	hkjob
hk_kjob:
	move.w	#hk.kjob,d0
hkjob:
	link	a6,#0               	; frame pointer
	movem.l	d2-d4/a2-a5,-(sp)
	move.w	d0,d4			; action
	bsr	hk_use              	; get hotkey system
	bne	hks_done            	; failed ...
	move.l	0(a3,d4.w),a4       	; our vector
	jsr	(a4)
	bra	hks_finish

hk_thing:
	dc.w	6
	dc.b	'Hotkey'

	end

