/*
 * Test / Demo for the Thing part of the hotstuff library
 *
 * The first time the program is run, it create a shared data thing
 * called "Test Stuff".
 *
 * The second time it's run, it USEs "Test Stuff", then creates a child
 * job that modifies the shared data, lists the jobs using the data
 * and then tidies up, removing the shared thing.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <qdos.h>
#include <qlib.h>
#include <pd/thing.h>

extern char *_sys_var;

/*
 * We use this structure to describe the our shared data thing
 * the flag and type fields are mandatory QJump fields. We have
 * 32 bytes for our data (this is a trivial example :-)
 */

typedef struct
{
    int  flag;
    int  type;
    char text[32];
} TEST_DATA, *PTEST_DATA;


#define TEST_ID  0x302e3031         /* 0.01 */

/*
 * Let's just have one of these
 */
 
static char *Stuff = "Test Stuff";

/*
 * Here we define our simple thing
 */
 
PTHG_LINK AddThing(char *name, char *txt, PTHG_LINK *p)
{
    int m,n;
    int size;
    int sts;
    
    n = strlen(name);

    /*
     * size = size of a linkage block plus the length of the name, and is
     * rounded to the next even number if necessary
     */
     
    size = sizeof(THG_LINK) + ((n + 1) & ~1);

    /*
     * Make one allocation for both linkage block and data
     * Owned by job zero, so it says when the program terminates
     * If the linkage is owned by the job, QDOS will tidy up for us 
     */
     
    *p = (PTHG_LINK) thg_malloc ((m = (size + sizeof(TEST_DATA))), 0);

    printf("Length is %d, size %d request %d\n", n ,size, m);

    if(*p)
    {
        /*
         * Got some memory, so define our thing. We have to set the id
         * the name (and length) and thing address in the linkage block
         * and the header info in the thing itself. We also preset the
         * data in the thing.
         */
         
        (*p)->namelen = n;
        memcpy(((*p)+1), name, n);
        (*p)->verid = TEST_ID;

        /*
         * This is where the thing starts.
         */
         
        (*p)->thing = ((char *)(*p) + size);

        ((PTEST_DATA) (*p)->thing)->flag = THH_FLAG;
        ((PTEST_DATA) (*p)->thing)->type = THT_DATA;
        strcpy( ((PTEST_DATA) (*p)->thing)->text, txt);
        sts = sms_lthg(*p);

        if(sts != 0)
        {
            thg_free(*p);
            *p = NULL;
        }
    }
    return sts;
}

/*
 * Sharer is a little concurrent job that shares our thing. We synchronise
 * usage by using address [31] in the shared data as a flag.
 */ 

int Sharer()
{
    int vers;
    PTHG_LINK p = NULL;    
    void *q;
    int sts;
    
    /*
     * Try to use the thing
     */
     
    sts = sms_uthg(Stuff, -1, -1, NULL, &(void *)p, &vers, &q);
    if(sts == 0)
    {
        /*
         * OK, we can use it, so change the data
         */
        strcpy(((PTEST_DATA)q)->text, "It's a wonderful thing");

        /*
         * Set the flag to show the data has changed
         */

        ((PTEST_DATA)q)->text[31] = 1;

        /*
         * Wait until the parent has done his (her) bit
         */

        while(((PTEST_DATA)q)->text[31])
        {
            mt_susjb(-1, 1, NULL);
        }

        /*
         * Tell QDOS we're done with the thing
         */

        sms_fthg(Stuff, -1, 0, NULL, NULL, NULL);

        /*
         * And also tell the parent job
         */

        ((PTEST_DATA)q)->text[31] = 1;
    }

    return 0;
}

/*
 * The real work starts here
 */

main(int ac, char **av)
{
    PTHG_LINK p = NULL;    
    long vers = 0;
    void *q;
    int sts;
    int jid;

    
    /*
     * See if we can use the thing
     */

    sts = sms_uthg(Stuff, -1, -1, NULL, &(void *)p, &vers, &q);

    if(sts == 0)
    {
        /* We already have a thing from the last invocation of the
         * program, so lets see if we can (ab)use it.
         */

        printf("use gives %d, vers is %4.4s addr %x %x\n", sts, &vers, q, p);
        getchar();

        if(sts == 0 && q)
        {
            /*
             * That was OK, so show header (THG%) and the type
             * (should be 2) and the data
             */
             
            printf("The text is (%4.4s %d) : %s\n",
                        q, *((int *)q + 1), ((PTEST_DATA)q)->text);

            /*
             * Set the flag so the (to be conceived) child can change
             * the data.
             */
             
            ((PTEST_DATA)q)->text[31] = 0;
        }

        /*
         * Now start the child job
         */
         
	jcall(Sharer, "Share a thing", 16, 1024, 0);

	/*
	 * Idle time while the child changes the thing
	 */
	 
	while(((PTEST_DATA)q)->text[31] == 0)
	{
	    mt_susjb(-1, 1, NULL);
	}

        /*
         * How see what's there
         */
         
	printf("The shared thing is now : %s\n", ((PTEST_DATA)q)->text);

        /*
         * Now see who is using the thing (should be us and the kid)
         * We prime sms_nthu by ensuring the pointer is NULL and
         * ignoring the first returned jobid
         */
         
        p = NULL;
	sms_nthu(Stuff, &p, &jid);

	/*
	 * Now loop through all users, displaying Job Ids and name
	 */
	 
	while(p)
	{
	    short *jadr;
	    int  jp;
	    
	    sms_nthu(Stuff, &p, &jid);
	    printf("The user is %4d %4d ", jid & 0xffff, jid >> 16);
            mt_jinf(&jid, &jid, &jp,  & (char *)jadr);

            if(*(jadr + 3) == 0x4afb)
            {
                printf("%-.*s\n", *(jadr+4), (jadr+5));
            }
            else
            {
                puts("no job name");
            }
	}

        puts("That's all");
        
        /*
         * Show the child it can terminate
         */
         
        ((PTEST_DATA)q)->text[31] = 0;

        /*
         * And wait until its done (ensures we can remove the thing)
         */
         
	while(((PTEST_DATA)q)->text[31] == 0)
	{
	    mt_susjb(-1, 1, NULL);
	}
        fputs("OK ? ", stdout);
        getchar();
      
        /*
         * Tell QDOS we're done with it
         */

        sts = sms_fthg(Stuff, -1, 0, NULL, NULL, NULL);
        printf("free gives %d\n", sts);
        getchar();

        /*
         * Now remove this troublesome thing
         */
         
        sts = sms_rthg(Stuff);
        printf("remove gives %d\n", sts);
    }
    else if(sts == ERR_NF)
    {
        sts = AddThing (Stuff, "What thing is this ?", &p);
        printf("Add returns %d (%x)\n", sts, p);
    }

    if(sts == ERR_NI)
    {
        poserr("No Thing system");
    }
    else
    {
        fputs("Press ENTER ", stdout);
        getchar();

        puts("This is the nthg list");

        for (q = NULL; sts = sms_nthg(q, &p), p;)
        {
            char num[16];
            char *np;
            printf("%4d %08x", sts, p);
            if((PTEST_DATA) q = p->thing)
            {
                int n;
                n = ((PTEST_DATA)q)->type;

                /*
                 * If its not the THING thing, clear out the
                 * list bit
                 */
                
                if(n > 0)
                {
                    n &= ~THT__LST;
                }
                np = itoa(num,  n);
            }
            else
            {
                np = "(None)";
	    }  
            /*
             * And show name and type
             */
            printf(" %-20.*s %s\n", p->namelen, (p+1), np);
	}
    }
    return 0;    
}
